/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il	    */
/* License: GNU General Public License                              */

/*
 * $Id: test.cpp,v 1.3 2002/07/18 21:54:25 mulix Exp $
 */

/* This is a program used to test the config parser.
 * @author: mulix@actcom.co.il
 */

#include <iostream>
#include <fstream>
#include <list>
#include <cstdio>

extern"C"{
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h> //for stat()
#include <sys/types.h> //for uid_t
#include <getopt.h> //for getopt 
}

#include "sct_ctrl_lib/sct_ctrl_lib.h"
#include "sct_rules/rule_engine_public.h"
#include "rule.h"
#include "logger.h"
#include "parser.h"
#include "syscalls.h"
#include "tokens_to_types.h"
#include "sstreamwrap.h"

using namespace std; 

const char DEFAULT_CONFIG_FILE_NAME[] = "sct_example.conf";
const char DEFAULT_SYSCALLS_INFO_LOCATION[] = "../sct_rules_module/syscalls.dat";

const char* progname; 

/* forward declarations */
bool load_config_file(const char * config_file_name,
		      sstreamwrap& data);
bool parse_config_file(sstreamwrap& data,
		       list<rule>& rules, 
		       const string& config_file_name);

#ifdef SCT_CONFIG_DEBUG
int debug = 1;
#else /* SCT_CONFIG_DEBUG is not defined */
int debug = 0;
#endif

void 
print_rule_list(syscall_rules_t* rules)
{
    assert(rules); 

    for ( ; rules; rules = rules->next) {
	char sid[32]; 
	sct_get_syscall_id_as_str(rules->syscall_id, sid, sizeof(sid)); 
	cout << "syscall: " << sid 
	     << ", " << (rules->when == RW_BEFORE? "before" : "after") 
	     << " rules:" << endl; 
	sct_print_tracking_rules(rules->list); 
    }
}

void 
free_rule_list(syscall_rules_t* rules)
{
    int first = 1; /* we dont free the head of the list */ 

    assert(rules); 

    for ( ; rules; rules = rules->next) {
	sct_delete_tracking_rule_list(rules->list); 
	rules->list = NULL;
	if (!first)
	    free(rules); 
	first = 0; 
    }
}

int main(int argc, char ** argv)
{
    const char* config_file_name = DEFAULT_CONFIG_FILE_NAME;
    sstreamwrap data;

    if (!load_config_file(config_file_name, data)){
	fprintf(stderr,
		"Error: failed to read the config file '%s'. quitting.\n", 
		config_file_name);
	return 1;
    }
    //the list of rules
    list<rule> rules;

    if (!parse_config_file(data,
			   rules,
			   config_file_name)){
	fprintf(stderr, "Error: failed to parse the config file. quitting.\n");
	return 1;
    }

    cout << "parsed rule list:" << endl;
    list<rule>::const_iterator it = rules.begin();
    for (; it != rules.end(); ++it){
	    cout << *it << endl;
    }

    return 0;
}


bool load_config_file(const char * config_file_name,
		      sstreamwrap& data)
{
    //first of all, make some sanity checks on the file

    //preferably, that user will be root, but we relax that requirement for
    //development
    struct stat stat_info;
    if (stat(config_file_name, &stat_info)==-1){
	fprintf(stderr,
		"load_config_file: stat '%s' failed, with error '%s'\n",
		config_file_name, strerror(errno));
	return false;
    }

    //the config file is a regular file
    if (!S_ISREG(stat_info.st_mode)){
	fprintf(stderr,
		"load_config_file: file '%s' is not a regular file\n", 
		config_file_name);
	return false;
    }

    //now, that we've verified the file is usable by us, let's load
    //its contents into a buffer.
    ifstream in(config_file_name);
    in>>data.rdbuf();
    return true;
}

bool 
parse_config_file(sstreamwrap& data,
		  list<rule>& rules,
		  const string& config_file_name)
{
    /* maps from string to enumerated types     */
    /* must be done here since the syscalls.dat */
    /* parser uses them as well                 */
    tokens::init();
    
    // load the syscall information file
    if (!syscalls::load(DEFAULT_SYSCALLS_INFO_LOCATION)) {
	fprintf(stderr,
		"Error: failed to find syscall information "
		"file at '%s'.\n", DEFAULT_SYSCALLS_INFO_LOCATION);
	return false;
    }
    
    //create a new parser
    parser p(data, config_file_name);
    //start parsing the config data
    return p.parse(rules);
}
