/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU Lesser General Public License                       */

/*
 *$Id: logging.c,v 1.7 2002/08/23 17:38:39 mulix Exp $
 */

#include "rule_engine_public.h"
#include "rule_engine_private.h"
#include "membuf.h"
#include "filter_vars_public.h"
#include "archdep.h"

/* maximum size of a log record. */
#define MAX_LOG_RECORD	1024

/* types of logging macros, used when logging syscall invocations. */
typedef enum {
	LM_PID = 0,
	LM_UID = 0,
	LM_EUID = 0
} log_macro_t;

/* maximum number of supported logger callbacks. */
#define	MAX_LOGGER_CALLBACKS	5

/* logger callbacks array. */
struct logger_callback logger_callbacks[MAX_LOGGER_CALLBACKS];
static int num_logger_callbacks = 0;

sct_membuf_t* log_dev_membuf; 

/* find the registered logger calback function, that handles  */
/* logging a given type of spec-char. returns the callback ID */
/* if one was registered, '-1' otherwise.                     */
int get_logger_callback_index(char spec_char)
{
	int i;

	for (i=0; i < num_logger_callbacks; i++) {
		if (logger_callbacks[i].spec_char == spec_char)
			break;
	}

	return (i < num_logger_callbacks ? i : -1);
}

/* register the given logger calback function, that handles */
/* logging a given type of spec-char                        */
void set_logger_callback(char spec_char, logger_callback_t cb, void* cb_data)
{
	int i;

	for (i=0; i < num_logger_callbacks; i++) {
		if (logger_callbacks[i].spec_char == spec_char)
			break;
	}

	if (i == num_logger_callbacks) {
		if (i >= MAX_LOGGER_CALLBACKS) 
			return; 

		num_logger_callbacks++;
	}
	logger_callbacks[i].spec_char = spec_char;
	logger_callbacks[i].cb = cb;
	logger_callbacks[i].cb_data = cb_data;
}

/* stringify the given string value, into the given buffer.           */
/* return 0 on success, -ERRNO otherwise                              */
static int tracker_stringify_string(const char* string, sct_membuf_t* m, 
				    int add_quotes)
{
	int rc; 
	char quote = '"'; 

	/* we want to output '"string"' */
	if (add_quotes) {
		if ((rc = sct_membuf_add_chunk(m, &quote, 1)) < 0)
			return rc; 
	}
	
	if (string) {
		if (strlen(string) < sct_membuf_free_for_write(m)) {
			if ((rc = sct_membuf_add_chunk(m, string, strlen(string))) < 0)
				return rc; 
		} else { /* try to output 100 bytes or less f the string if we can */ 
			int cnt = sct_min(sct_membuf_free_for_write(m) - strlen("..."), 100); 
			if (cnt > 0) {
				if ((rc = sct_membuf_add_chunk(m, string, cnt)) < 0)
					return rc; 
				if ((rc = sct_membuf_add_chunk(m, "...", strlen("..."))) < 0)
					return rc; 
			}
		}
	} else { /* the string was null */
		if ((rc = sct_membuf_add_chunk(m, "(nil)", strlen("(nil"))) < 0)
			return rc; 
	}
	/* emit the trailing '"'. */
	if (add_quotes) {
		if ((rc = sct_membuf_add_chunk(m, &quote, 1)) < 0)
			return rc; 
	}

	return 0;
}

/* stringify the given unsigned numeric value, into the given buffer, */
static int tracker_stringify_unum(unsigned long num, sct_membuf_t* m)
{
	int rc; 
	char sbuf[32];
    
	sprintf(sbuf, "%lu", num);
	rc = sct_membuf_add_chunk(m, sbuf, strlen(sbuf)); 

	return rc; 
}

/* stringify the given pointer value, into the given buffer,   */
static int tracker_stringify_pointer(void* p, sct_membuf_t* m)
{
	char sbuf[32];
	int rc = 0;
    
	rc = sprintf(sbuf, "%p", p);
	rc = sct_membuf_add_chunk(m, sbuf, strlen(sbuf)); 

	return rc; 
}

/* stringify the given unsigned numeric value, into the given buffer, */
static int tracker_stringify_num(long num, sct_membuf_t* m)
{
	int rc; 
	char sbuf[32];
    
	sprintf(sbuf, "%ld", num);
	rc = sct_membuf_add_chunk(m, sbuf, strlen(sbuf)); 

	return rc; 
}

/* stringify the given unsigned numeric value, into the given buffer, */
static int tracker_stringify_num64(long long num, sct_membuf_t* m)
{
	int rc; 
	char sbuf[32];
    
	sprintf(sbuf, "%lld", num);
	rc = sct_membuf_add_chunk(m, sbuf, strlen(sbuf)); 

	return rc; 
}

/* stringify the given variable into the given buffer.                */
static int tracker_stringify_var_node(sct_membuf_t* m, 
				      sct_filter_node_t *var_val)
{
	int rc = 0;
	char* string = NULL;

	if (!var_val)
		return tracker_stringify_string("<NULL>", m, 0); 
	if (var_val->type != NT_CONST) {
		PPRINT("eek, internal incosistency, %s got a filter node of type %d\n", 
		       __FUNCTION__, var_val->type); 
		return 0; 
	}
		
	switch (var_val->data_type) {
	case DT_USHORT:
		rc = tracker_stringify_unum(CONST_USHORT(var_val), m);
		break;
	case DT_SHORT:
		rc = tracker_stringify_num(CONST_SHORT(var_val), m);
		break;
	case DT_UINT:
		rc = tracker_stringify_unum(CONST_UINT(var_val), m);
		break;
	case DT_INT:
		rc = tracker_stringify_num(CONST_INT(var_val), m);
		break;
	case DT_ULONG:
		rc = tracker_stringify_unum(CONST_ULONG(var_val), m);
		break;
	case DT_LONG:
		rc = tracker_stringify_num(CONST_LONG(var_val), m);
		break;
	case DT_LONG_LONG: 
		rc = tracker_stringify_num64(CONST_LONG_LONG(var_val), m); 
		break; 
	case DT_STRING:
		string = CONST_STRING(var_val);
		rc = tracker_stringify_string(string, m, 1);
		break;
	case DT_BUFFER:
		string = sct_buffer_to_string(CONST_BUFFER(var_val), 
					      BUFFER_SIZE(var_val)); 
		rc = tracker_stringify_string(string, m, 1); 
		mem_free(string); 
		break; 
	default:
		break;
	}
    
	return rc; 
}

static int tracker_stringify_var(sct_membuf_t* m, sct_var_type_t var)
{
	return tracker_stringify_var_node(m, get_var(var));
}

static int tracker_stringify_var_indexed(sct_membuf_t* m, 
					 sct_var_type_t var, 
					 int var_index)
{
	sct_filter_node_t* t = get_var_indexed(var, var_index); 
	return tracker_stringify_var_node(m, t);
}

/* stringify the parameters of the syscall. */
static int tracker_stringify_params(sct_membuf_t* m,
				    const char* param_spec,
				    va_list ap)
{
	char c;
	unsigned char uc;
	long l;
	unsigned long ul;
	long long ll; 
	int rc = 0;
	void* v, *p;

	if (param_spec) {
		int next_is_unsigned = 0;
		char* string = NULL; 
		char this_char = 0; 
		char next_char = 0; 
		int is_string; /* strings require a bit of special handling */ 
		int skip_params = 0; /* are we in 'parameters skipping' mode? */
		int logger_cb_id = -1;

		while (*param_spec) {
			this_char = *param_spec++;
			/* handle logger callbacks first. */
			logger_cb_id = get_logger_callback_index(this_char);
			if (logger_cb_id > -1) {
				void* p = va_arg(ap, void*);
				void* p_data = va_arg(ap, void*);
				/* PPRINT("calling logger_callback with struct data %p\n", p_data); */ 
				rc = (*logger_callbacks[logger_cb_id].cb)(
					this_char,
					m, 
					p,
					p_data,
					logger_callbacks[logger_cb_id].cb_data);
				goto tracker_stringify_params_next_char;
			}

			is_string = (this_char == 's');

			switch (this_char){
			case ' ': break; /* skip spaces */
			case 'u': next_is_unsigned = 1; break;
			case 'c':
				if (next_is_unsigned) {
					uc = (unsigned char)va_arg(ap, unsigned int);
					if (!skip_params) 
						sct_membuf_add_chunk(m, &uc, 1); 
					
					next_is_unsigned = 0;
				} else {
					c = (char)va_arg(ap, int);
					if (!skip_params)
						sct_membuf_add_chunk(m, &c, 1); 
				}
				break;
			case 'h':
				if (next_is_unsigned) {
					ul = (unsigned short)va_arg(ap, unsigned int);
					if (!skip_params)
						rc = tracker_stringify_unum(ul, m); 
				
					next_is_unsigned = 0;
				} else {
					l = (short)va_arg(ap, int);
					if (!skip_params) 
						rc = tracker_stringify_num(l, m);
				}
				break;
			case 'i':
				if (next_is_unsigned) {
					ul = va_arg(ap, unsigned int);
					if (!skip_params) 
						rc = tracker_stringify_unum(ul, m);

					next_is_unsigned = 0;
				} else {
					l = va_arg(ap, int);
					if (!skip_params) 
						rc = tracker_stringify_num(l, m);
				}
				break;
			case 'l':
				if (next_is_unsigned) {
					ul = va_arg(ap, unsigned long);
					if (!skip_params) 
						rc = tracker_stringify_unum(ul, m);

					next_is_unsigned = 0;
				} else {
					l = va_arg(ap, long);
					if (!skip_params) 
						rc = tracker_stringify_num(l, m);

				}
				break;
                        case 'L': 
				ll = va_arg(ap, long long); 
				if (!skip_params)
					rc = tracker_stringify_num(ll, m); 
				break; 
			case 'p':
				p = va_arg(ap, void*);
				if (!skip_params) 
					rc = tracker_stringify_pointer(p, m);

				break;
			case 'P': /* indexed var */
				l = va_arg(ap, int);
				if (!skip_params) 
					rc = tracker_stringify_var_indexed(m, VT_PARAM, l);

				break;
			case 'B': /* indexed var */ 
				l = va_arg(ap, int);
				if (!skip_params)
					rc = tracker_stringify_var_indexed(m, VT_PARAM, l);

				break; 
			case 'T': /* type-name */
				string = va_arg(ap, char*);
				if (!skip_params)
					rc = tracker_stringify_string(string, m, 0);
				break;
			case 's': /* string */
			case 'S': /* struct */
				string = va_arg(ap, char*);
				if (!skip_params) {
					rc = tracker_stringify_string(string,
								      m, 
								      is_string);

					/* for a string, we want to output '"string"' */
					if (is_string) {
						/* for strings, we are done */
						break;
					}
					/* for structures, the next arg is a pointer to the */
					/* _user space_ structure we only care if it's NULL */
					/* or not, and we never ever dereference it.        */
					v = va_arg(ap, void*);
					if (v == NULL)
						skip_params = 1;
				}
				break;
				
			case '{':
				rc = sct_membuf_add_chunk(m, "{", 1); 
				break;
			case '}':
				rc = sct_membuf_add_chunk(m, "}", 1); 
				if (skip_params)
					skip_params = 0;
				break;
			case 'v':
				if (!skip_params) 
					rc = sct_membuf_add_chunk(m, "void", strlen("void")); 
				break;
			default:
				break;
			}

		tracker_stringify_params_next_char:
			/* if we just printed a value, print a comma after it */
			next_char = *param_spec;
			if (logger_cb_id > -1 && next_char && next_char != '}') {
				if (!skip_params) 
					sct_membuf_add_chunk(m, ", ", strlen(", ")); 

			} else if ((this_char != ' ') && (this_char != 'u') && 
				   (this_char != 'S') && (this_char != 'T') && 
				   (this_char != '{') && (next_char) && next_char != '}') {  
				if (!skip_params) 
					sct_membuf_add_chunk(m, ", ", strlen(", ")); 
			}
		}
		if (rc < 0)
			return rc; 
	}
	return 0; 
}


/* stringify the parameters of the syscall. */
int sct_tracker_stringify_params(sct_membuf_t* m, 
				 const char* param_spec,
				 ...)
{
	int rc;
	va_list ap;
	va_start (ap, param_spec);

	rc = tracker_stringify_params(m, param_spec, ap);
	va_end(ap);

	return rc;
}

/* perform macro translation into the given buffer with the given size. */
/* returns 0 on success or -ERRNO                   .                   */
static int tracker_translate_macro(int rule_id, 
				   int syscall_id, 
				   const char* syscall_name, 
				   const char* macro, 
				   sct_membuf_t* m, 
				   const char* param_spec, 
				   va_list ap)
{
	if (!strcmp(macro, "ruleid"))
		return tracker_stringify_num(rule_id, m);

	if (!strcmp(macro, "sid"))
		return tracker_stringify_num(syscall_id, m);

	if (!strcmp(macro, "sname")) 
		return sct_membuf_add_chunk(m, syscall_name, strlen(syscall_name)); 
		
	if (!strcmp(macro, "params"))
		return tracker_stringify_params(m, param_spec, ap);
		
	if (!strcmp(macro, "pid"))
		return tracker_stringify_var(m, VT_PROC_PID);

	if (!strcmp(macro, "uid"))
		return tracker_stringify_var(m, VT_PROC_UID);

	if (!strcmp(macro, "euid")) 
		return tracker_stringify_var(m, VT_PROC_EUID);

	if (!strcmp(macro, "suid"))
		return tracker_stringify_var(m, VT_PROC_SUID);

	if (!strcmp(macro, "gid"))
		return tracker_stringify_var(m, VT_PROC_GID);

	if (!strcmp(macro, "egid"))
		return tracker_stringify_var(m, VT_PROC_EGID);
	
	if (!strcmp(macro, "sgid"))
		return tracker_stringify_var(m, VT_PROC_SGID);

	if (!strcmp(macro, "comm"))
		return tracker_stringify_var(m, VT_PROC_COMM);

	if (!strcmp(macro, "ppid"))
		return tracker_stringify_var(m, VT_PROC_PPID);

	if (!strcmp(macro, "pcomm"))
		return tracker_stringify_var(m, VT_PROC_PCOMM);

	if (!strcmp(macro, "retval"))
		return tracker_stringify_var(m, VT_RETVAL);

	PPRINT("%s: unknown macro '%s'\n", __FUNCTION__, macro); 
	return 0; 
}

/* log the syscall that matched one of the rules in 'rules' */
void tracker_do_logging(sct_membuf_t* mbuf, 
			int rule_id, 
			int syscall_id, 
			const char* syscall_name, 
			const char* log_format, 
			const char* param_spec, 
			va_list ap)
{
	char macro[MAX_LOG_RECORD+1];
	const char* pf = log_format;
	char const* ps; 
	int mlen; 
	sct_membuf_t* tmpbuf; 
	int using_printk = 0; 

	if (!mbuf) {
		if (!(tmpbuf = sct_membuf_create(MAX_LOG_RECORD + 1, NULL, NULL, 
						 SCT_MEMBUF_EXACT_MATCH))) {
			PPRINT(KERN_WARNING "failed to allocate membuf, quitting\n"); 
			return; 
		}
		mbuf = tmpbuf; 
		using_printk = 1; 
	}
	
	while (*pf) {
		if (*pf != '%') {
			sct_membuf_add_chunk(mbuf, pf, 1); 
			pf++; 
			continue; 
		}
		pf++; 
		if (!*pf)
			break; 

		if (*pf == '%') {
			/* translate '%%' to '%'. */
			sct_membuf_add_chunk(mbuf, "%", 1); 
			pf++; 
			continue; 
		} else {
			/* translate macro name into value. */
			ps = pf; 
			/* parse out the macro name - from the current char
			   until the next non alphabetic character */
			while (isalpha(*ps++))
				;
			ps--;
			mlen = sct_min(ps - pf, sizeof(macro) - 2);
			strncpy(macro, pf, mlen);
			macro[mlen] = '\0';	
			tracker_translate_macro(rule_id,
						syscall_id,
						syscall_name,
						macro,
						mbuf,
						param_spec,
						ap);
			pf = ps;
			continue;
		}
	}

	if (using_printk) {
		PPRINT(KERN_NOTICE "%s", sct_membuf_get_data(mbuf));
		sct_membuf_destroy(mbuf); 
	}
}

int sct_register_log_membuf(sct_membuf_t* log_mbuf)
{
	log_dev_membuf = log_mbuf; 
	return 0; 
}

void sct_deregister_log_membuf(void)
{
	log_dev_membuf = NULL; 
}
