/* This file is part of syscalltrack, a GNU/Linux kernel module and */
/* user space utilities for logging or modifying any system call    */
/* invocation.                                                      */
/*                                                                  */
/* Copyright (C) 2000-2002 Muli Ben-Yehuda, mulix@actcom.co.il      */
/* License: GNU General Public License                              */

/*
 * $Id: log_dev.c,v 1.5 2002/08/04 03:33:49 mulix Exp $
 */ 

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/version.h>      /* module things */
#include <linux/module.h>       /* module things */
#include <linux/kernel.h>       /* printk() */
#include <linux/errno.h>        /* error codes */
#include <linux/types.h>        /* size_t */
#include <linux/proc_fs.h>      /* /proc stuff */ 
#include <asm/atomic.h>         /* test_and_set_bit() */ 
#include <asm/uaccess.h>        /* copy_to/from_user */

#include "sct_compat.h"
#include "log_dev.h"
#include "sct_rules_private.h"
#include "module_interface.h"
#include "util.h"

#define SCT_LOG_MAJOR 0   /* dynamic major by default */

#define SCT_RING_BUFFER_SIZE 32 * 1024

/* the major number we use */
static int log_major = SCT_LOG_MAJOR;

/* 1 is the device file is in use, 0 if we're free at the moment */ 
volatile long log_dev_in_use; 

sct_log_dev_t log_dev; 

extern struct file_operations log_fops; 

int init_log_dev_file(sct_log_dev_t** pdev)
{
	int retval;
	
	sct_log_dev_t* dev; 
	
	dev = *pdev = &log_dev;

	retval = register_chrdev(log_major, "sct_log", &log_fops);
	if (retval < 0)
		goto unset_dev; 

	if (log_major == 0) 
		log_major = retval; /* dynamic */

	retval = -ENOMEM; 
	if (!(dev->ring = sct_membuf_create(SCT_RING_BUFFER_SIZE, NULL, sct_cp_to_user, 
					    SCT_MEMBUF_DROP_OLD))) 
		goto unregister_chrdev; 

#if 0 /* don't use the log device by default */ 
	if ((retval = sct_register_log_membuf(dev->ring)))
		goto deallocate_ring; 
#endif 
	
	retval = 0; 	

	init_waitqueue_head (&dev->wait);

	return retval; 

#if 0 
 deallocate_ring:	
	sct_membuf_destroy(dev->ring); 
	dev->ring = NULL; 
#endif 

 unregister_chrdev:
	unregister_chrdev(log_major, "sct_log");

 unset_dev:
	*pdev = NULL; 

	return retval; 
}

void destroy_log_dev_file(sct_log_dev_t* dev)
{
	if (!dev)
		BUG(); 

	/* in case it was registered */ 
	sct_deregister_log_membuf(); 
	
	if (dev->ring) {
		sct_membuf_destroy(dev->ring); 
		dev->ring = NULL; 
	}
	
	unregister_chrdev(log_major, "sct_log");
}

/*
 * Open and close
 */
static int log_open(struct inode *inode, struct file *filp)
{
	/* only one process make talk to us. this is ok for the configuration   */
	/* device file, but we'll have to redo this to support multiple readers */
	/* for the logging device file */ 
	if (test_and_set_bit(0, &log_dev_in_use))
		return -EBUSY; /* already open */ 

	if (!capable(CAP_SYS_PTRACE)) {
		clear_bit(0, &log_dev_in_use); 
		return -EPERM;
	}
	
	MOD_INC_USE_COUNT;

	/* and use filp->private_data to point to the device data */
	filp->private_data = &log_dev;
	
	return 0;          /* success */
}

static int log_release(struct inode *inode, struct file *filp)
{
	if (!log_dev_in_use)
		BUG(); 

	clear_bit(0, &log_dev_in_use); 

	MOD_DEC_USE_COUNT;

	return 0;
}

/* some of the read code is adapted from systrace.c, at */ 
/* http://www-sor.inria.fr/~patarin/systrace.c */ 
static int log_read(struct file *filp, char *buf, size_t count,
		    loff_t *f_pos)
{
	sct_log_dev_t *dev = (sct_log_dev_t*)filp->private_data;

	ssize_t ret; 

	ret = -ERESTARTSYS;
	if (down_interruptible(&tracker_sem))
		goto out_nolock;

	/* nothing to read */
	while (!sct_membuf_available_for_read(dev->ring)) { 
		ret = -EAGAIN;
		if (filp->f_flags & O_NONBLOCK)
			goto out;
		up(&tracker_sem);
		interruptible_sleep_on(&dev->wait);
		ret = -ERESTARTSYS;
		if (down_interruptible(&tracker_sem))
			goto out_nolock;

		ret = -ERESTARTSYS;
		if (signal_pending(current)) 	/* a signal arrived */
			goto out; 		/* tell the fs layer to handle it */
		/* otherwise loop */
	}

	ret = sct_membuf_get_chunk(dev->ring, buf, count);

 out:
	up (&tracker_sem);
 out_nolock:
	return ret;
}

static int log_write(struct file *filp, const char *buf, size_t count,
		     loff_t *f_pos)
{
	return -EBADF; 
}

static int log_ioctl(struct inode *inode, struct file *filp,
		     unsigned int cmd, unsigned long arg)
{
	int ret;
	sct_log_dev_t* dev = (sct_log_dev_t*)filp->private_data;

	printk(KERN_NOTICE "log_ioctl called with cmd %d\n", cmd); 
	
	ret = -ERESTARTSYS;
	if (down_interruptible(&tracker_sem))
		goto out_nolock; 

	ret = 0;
	
	switch (cmd) {
	case CTL_IOC_DROP_OLD: 
		sct_membuf_drop_old(dev->ring); 
		break; 

	case CTL_IOC_DROP_NEW:
		sct_membuf_drop_new(dev->ring); 
		break; 

	case CTL_IOC_FLUSH: 
		sct_membuf_clear(dev->ring); 
		break;

	case CTL_IOC_ENABLE_DEV: 
		sct_register_log_membuf(dev->ring); 
		break; 
		
	case CTL_IOC_DISABLE_DEV:
		sct_deregister_log_membuf(); 
		break; 
		
	default:
		ret =  -EINVAL;
		break;
	}
	up(&tracker_sem);
	
 out_nolock:
	return ret;
}

/*
 * The fops
 */

struct file_operations log_fops = {
	MODULE_OWNER(THIS_MODULE)
	.open = log_open,
	.release = log_release, 
	.read = log_read, 
	.write = log_write, 
	.ioctl = log_ioctl
};

void wake_up_logger(void)
{
	wake_up_interruptible(&log_dev.wait);
}
